
----------------------------------------------------------------
--------------------     body  StringPack    -------------------
----------------------------------------------------------------

WITH text_io, calendar; USE text_io;
PRAGMA ELABORATE(text_io, calendar);

PACKAGE BODY StringPack IS

-------------- type VARSTRING and associated operations -------------
---------------------------------------------------------------------

FUNCTION  "&" (va: VARSTRING; vb: VARSTRING) RETURN VARSTRING IS
  v : VARSTRING := va;  vbl: INTEGER := vb.lg;
BEGIN
  IF v.lg + vbl > maxvarlg THEN
    vbl := maxvarlg - v.lg;
  END IF;
  v.str (v.lg + 1 .. v.lg + vbl) := vb.str (1 .. vbl);
  v.lg := v.lg + vbl;
  RETURN v;
END "&";

FUNCTION  "&" (va: STRING; vb: VARSTRING) RETURN VARSTRING IS
  v: VARSTRING;  valg: INTEGER := va'length;  vblg: INTEGER := vb.lg;
BEGIN
  IF valg > maxvarlg THEN
    valg := maxvarlg;
  END IF;
  IF valg+vblg > maxvarlg THEN
    vblg := maxvarlg-valg;
  END IF;
  v.lg := valg+vblg;
  v.str(1..valg) := va(va'FIRST..va'FIRST+valg-1);
  v.str(valg+1..v.lg) := vb.str(1..vblg);
  RETURN v;
END "&";

FUNCTION  "&" (va: VARSTRING; vb: STRING) RETURN VARSTRING IS
  v: VARSTRING := va;
  vbl: INTEGER := vb'length;
BEGIN
  IF v.lg + vbl > maxvarlg THEN
    vbl := maxvarlg - v.lg;
  END IF;
  v.str(v.lg+1..v.lg+vbl) := vb(vb'FIRST..vb'FIRST+vbl-1);
  v.lg := v.lg + vbl;
  RETURN v;
END "&";

FUNCTION "&" (va: VARSTRING; i:INTEGER) RETURN VARSTRING IS
BEGIN RETURN va & cvis(i); END "&";

FUNCTION "&" (i:INTEGER; va: VARSTRING) RETURN VARSTRING IS
BEGIN RETURN cvis(i) & va; END "&";

FUNCTION "&" (s:STRING; i: INTEGER) RETURN VARSTRING IS
BEGIN RETURN s & cvis(i); END "&";

FUNCTION "&" (i:INTEGER; s:STRING) RETURN VARSTRING IS
BEGIN RETURN cvis(i) & s; END "&";

FUNCTION "&" (va: VARSTRING; f:FLOAT) RETURN VARSTRING IS
BEGIN RETURN va & cvfs(f); END "&";

FUNCTION "&" (f:FLOAT; va: VARSTRING) RETURN VARSTRING IS
BEGIN RETURN cvfs(f) & va; END "&";

FUNCTION "&" (s:STRING; f:FLOAT) RETURN VARSTRING IS
BEGIN RETURN s & cvfs(f); END "&";

FUNCTION "&" (f:FLOAT; s:STRING) RETURN VARSTRING IS
BEGIN RETURN cvfs(f) & s; END "&";

FUNCTION equal (v: VARSTRING; s: STRING) RETURN BOOLEAN IS
BEGIN RETURN v.lg=s'length AND THEN v.str(1..v.lg)=s; END equal;

FUNCTION equal (s: STRING; v: VARSTRING) RETURN BOOLEAN IS
BEGIN RETURN v.lg=s'length AND THEN v.str(1..v.lg)=s; END equal;

FUNCTION equal (va: VARSTRING; vb: VARSTRING) RETURN BOOLEAN IS
BEGIN RETURN va.lg=vb.lg AND THEN va.str(1..va.lg)=vb.str(1..vb.lg); END equal;

FUNCTION  "+" (v: VARSTRING) RETURN STRING IS
BEGIN RETURN v.str (1 .. v.lg); END "+";

FUNCTION  "+" (v: STRING) RETURN VARSTRING IS
  vs: VARSTRING;
BEGIN
  vs.lg:= v'LENGTH;
  IF vs.lg > maxvarlg THEN
    vs.lg := maxvarlg;
  END IF;
  vs.str (1 ..vs.lg) := v (v'FIRST .. v'FIRST + vs.lg - 1);
  RETURN vs;
END "+";

FUNCTION "*" (n: INTEGER; c: CHARACTER) RETURN VARSTRING IS
  vs: VARSTRING := ((others => c), n);
BEGIN
  IF n > maxvarlg THEN
    vs.lg := maxvarlg;
  END if;
  IF n < 0 THEN
    vs.lg := 0;
  END IF;
  RETURN vs;
END "*";

FUNCTION slice (v: VARSTRING; from, to: INTEGER) RETURN VARSTRING IS
  left: INTEGER:= from; right: INTEGER := to; res: VARSTRING;
BEGIN
   IF left < 1 THEN
     left := 1;
   END IF;
   IF right > v.lg THEN
     right := v.lg;
   END IF;
   IF right < left THEN
     RETURN varnull;
   END IF;
   -- here we know:  1 <= left <= right < v.lg
   res.lg := right - left + 1;
   res.str(1..res.lg) := v.str(left..right);
   RETURN res;
END slice;

----------------------- conversion functions ------------------------
---------------------------------------------------------------------

FUNCTION  cvis (item  : INTEGER;
                width : INTEGER := 0;
                filler: CHARACTER := ' ';
                base  : POSITIVE := 10) RETURN VARSTRING IS
  s       : varstr := (others => filler);
  x       : INTEGER := item;
  mydigits: constant STRING := "0123456789ABCDEF";
  k,lg    : NATURAL; r : INTEGER;
  lower   : INTEGER := s'LAST - width +1;
BEGIN
  FOR i IN REVERSE s'RANGE LOOP
    r := x REM base;
    x := x / BASE;
    s(i) := mydigits(ABS r + 1);
    IF x = 0 THEN
      k := i;
      EXIT;
    END IF;
  END LOOP; -- digits for abs item are now in s(k..s'LAST).
  IF item < 0 THEN
    k := k - 1;
  END if;
  IF lower not in 1..k THEN
    lower := k;
  END if;
  IF item < 0 THEN
    IF filler = ' ' THEN
      s(k) := '-';
    ELSE
      s(lower) := ' ';
    END if;
  END if;
  lg := s'LAST-lower+1;
  s(1..lg) := s(lower..s'LAST);
  RETURN (s,lg);
END cvis;

FUNCTION  cvls (item  : LONG_INTEGER;
                width : INTEGER := 0;
                filler: CHARACTER := ' ';
                base  : POSITIVE := 10) RETURN VARSTRING IS
  s       : varstr := (others => filler);
  x       : LONG_INTEGER := item;
  b       : LONG_INTEGER := LONG_INTEGER(base);
  mydigits: constant STRING := "0123456789ABCDEF";
  k,lg    : NATURAL;
  r       : INTEGER;
  lower   : INTEGER := s'LAST - width + 1;
BEGIN
  FOR i IN REVERSE s'RANGE LOOP
    r := INTEGER(x REM b);
    x := x / b;
    s(i) := mydigits(ABS r + 1);
    IF x = 0 THEN
      k := i;
      EXIT;
    END IF;
  END LOOP; -- digits for abs item are now in s(k..s'last).
  IF item < 0 THEN
    k := k - 1;
  END IF;
  IF lower NOT IN 1..k THEN
    lower := k;
  END IF;
  IF item < 0 THEN
    IF filler = ' ' THEN
      s(k) := '-';
    ELSE
      s(lower) := ' ';
    END IF;
  END IF;
  lg := s'LAST-lower+1;
  s(1..lg) := s(lower..s'LAST);
  RETURN (s,lg);
END cvls;

FUNCTION  cvfs (item: FLOAT; fore: INTEGER:= 0; aft: INTEGER := 2)
     RETURN VARSTRING IS
  af: INTEGER := 0; lg, wholestart: INTEGER;
  iitem: long_INTEGER; f: FLOAT := abs item;
  flast: constant FLOAT := FLOAT(long_INTEGER'LAST);
  s: varstr := (others=>' '); pos: INTEGER := s'LAST + 1;
  decdigits: constant STRING(1..10) := "0123456789";
BEGIN
  -- scale AND round to desired precision.
  WHILE af < aft AND THEN f*10.0 < flast LOOP
     -- note the invariant: f = abs item * 10.0**af
     f := f * 10.0;
     af := af+1;
  END LOOP;
  IF f >= flast THEN -- avoid Meridian bug
    raise StringPack_error;
  END IF;
  iitem := long_INTEGER(f); -- may raise exception (not for Meridian)
  -- compute fractional part and add decimal point
  FOR i IN 1..af LOOP
    pos := pos -1;
    s(pos) := decdigits(INTEGER(iitem rem 10 + 1));
    iitem := iitem / 10;
  END LOOP;
  IF af > 0 THEN
    pos := pos -1;
    s(pos) := '.';
  END IF;
  -- add whole part
  wholestart := pos;
  LOOP
     pos := pos -1;
     s(pos) := decdigits(INTEGER(iitem rem 10 + 1));
     iitem := iitem / 10;
     EXIT WHEN iitem = 0;
  END LOOP;
  IF item < 0.0 THEN
    pos := pos -1;
    s(pos) := '-';
  END IF;
  IF fore > wholestart - pos THEN
    pos := wholestart - fore;
  END IF;
  IF pos < 1 THEN
    pos := 1;
  END IF;
  -- decimal representation is now in s(pos .. s'LAST). shift left.
  lg := s'LAST+1-pos;
  s(1..lg) := s(pos..s'LAST);
  RETURN(s,lg);
  EXCEPTION WHEN OTHERS => RAISE StringPack_error;
END cvfs;

  FUNCTION cvsi (item  : VARSTRING) RETURN INTEGER IS
  -- verbesserte Version von LEO, 21.11.1993
    limit      : NATURAL := item.lg;
    my_num     : INTEGER := 0;
    signum     : INTEGER := 1;
    i          : INTEGER := 1;
    rememdigit : BOOLEAN;
  BEGIN
    IF limit < i THEN
      RAISE StringPack_error;
    END IF;
    -- skip leading blanks
    WHILE item.str(i) = ' ' AND THEN i < limit LOOP
      i := i + 1;
    END LOOP;
    -- deal with sign, if any
    IF item.str(i) = '-' THEN
      signum := - 1;
    END IF;
    IF item.str(i) = '-' OR item.str(i) = '+' THEN
      i := i+1;
    END IF;
    -- deal with decimal literal
    rememdigit := false;
    WHILE i <= limit LOOP
      IF item.str(i) IN '0' .. '9' THEN
        my_num := my_num*10  - CHARACTER'pos('0') + CHARACTER'pos(item.str(i));
        rememdigit := true;
      ELSE
        EXIT WHEN NOT rememdigit OR ELSE item.str(i) /= '_';
        rememdigit := false;
      END IF;
      i := i + 1;
    END LOOP;
    IF NOT rememdigit THEN
      RAISE StringPack_error;
    END IF;
    -- wind up
    FOR j IN i + 1 .. limit LOOP
      IF item.str(j) /= ' ' THEN
        RAISE StringPack_error;
      END IF;
    END LOOP;
    RETURN my_num * signum;
  END cvsi;

FUNCTION cvsf(item: IN VARSTRING) RETURN FLOAT IS
  c, exponentsign : CHARACTER;
  exponent, after : INTEGER;
  i               : NATURAL;
  res             : FLOAT := 0.0;
  p10             : FLOAT;
  negativ         : BOOLEAN := FALSE;
BEGIN
  c   := item.str(1);
  IF (c = '+') THEN
    i := 2;
  ELSIF (c = '-') THEN
    i := 2;
    negativ := TRUE;
  ELSE
    i := 1;
  END IF;
  c := item.str(i);
  WHILE (c /= '.') LOOP
    IF (c > '9') OR (c < '0') THEN
      IF (i = item.lg) THEN
        RAISE StringPack_Error;
      ELSIF (c = '_') THEN
        i := i + 1;
        c := item.str(i);
      ELSE
        c := '.';
        i := i - 1;
      END IF;
    ELSE
      res := res * 10.0 + FLOAT(CHARACTER'POS(c) - CHARACTER'POS('0'));
      i := i + 1;
      c := item.str(i);
    END IF;
  END LOOP;
  after := 0;
  i := i + 1;
  c := item.str(i);
  WHILE ( i <= item.lg ) AND ( c /= 'E' ) LOOP
    IF (c = '_') THEN
      i := i + 1;
    ELSE
      IF ( c > '9' ) OR ( c < '0' ) THEN
        RAISE StringPack_Error;
      END IF;
      res := res * 10.0 + FLOAT(CHARACTER'POS(c) - CHARACTER'POS('0'));
      i := i + 1;
      after := after + 1;
    END IF;
    IF ( i <= maxvarlg) THEN
      c := item.str(i);
    END IF;
  END LOOP;
  IF (i <= item.lg) AND THEN (c = 'E') THEN
    i := i + 1;
    exponentsign := item.str(i);
    IF (exponentsign = '+') OR (exponentsign = '-') THEN
      i := i + 1;
    END IF;
    c := item.str(i);
    exponent := 0;
    WHILE (i <= item.lg) LOOP
      IF (c /= '_') THEN
        IF ( c > '9' ) OR ( c < '0' ) THEN
          RAISE StringPack_Error;
        END IF;
        exponent := exponent * 8 + exponent * 2
                    + INTEGER(CHARACTER'POS(c) - CHARACTER'POS('0'));
      END IF;
      i := i + 1;
      c := item.str(i);
    END LOOP;
    IF exponentsign = '-' THEN
      exponent := - exponent;
    END IF;
  ELSE
    exponent := 0;
  END IF;
  exponent := exponent - after;
  p10 := 1.0;
  FOR i IN 1 .. ABS(exponent) LOOP
    p10 := p10 * 10.0;
  END LOOP;
  IF negativ THEN
    res := - res;
  END IF;
  IF (exponent < 0) THEN
    RETURN (res / p10);
  ELSE
    RETURN (res * p10);
  END IF;
END cvsf;

FUNCTION cvlu (item: VARSTRING) RETURN VARSTRING IS
  v: VARSTRING := item;
  d: constant := CHARACTER'POS('a') - CHARACTER'POS('A');
BEGIN
  FOR i IN 1..v.lg LOOP
    IF v.str(i) IN 'a' .. 'z' THEN
      v.str(i) := CHARACTER'VAL (CHARACTER'POS(v.str(i)) - d);
    END if;
  END LOOP;
  RETURN v;
END cvlu;

FUNCTION cvul (item: VARSTRING) RETURN VARSTRING IS
  v: VARSTRING := item;
  d: constant := CHARACTER'POS('a')- CHARACTER'POS('A');
BEGIN
  FOR i IN 1..v.lg LOOP
    IF v.str(i) IN 'A' ..'Z' THEN
      v.str(i) := CHARACTER'VAL (CHARACTER'POS(v.str(i)) + d);
    END if;
  END LOOP;
  RETURN v;
END cvul;

----------------------  date and time -------------------------
---------------------------------------------------------------

FUNCTION timestamp RETURN VARSTRING IS
  y,m,d,t,hrs,min,sec,sec100 : NATURAL; s: duration;
BEGIN
  calendar.split (calendar.clock, y, m, d, s);
  -- note the invariant
  --     hrs*3600 + minutes*60 + s = seconds since midnight
  -- in the following computations

  -- compute hrs in this day and subtract from seconds
  hrs := INTEGER(s/3600);
  s := s - hrs * duration(3600.0);
  IF s < duration(0.0) THEN
    hrs := hrs-1;
    s := s+duration(3600.0);
  END IF;

  -- compute minutes in this hour and subtract from seconds
  min := INTEGER(s/60);
  s := s - min * duration(60.0);
  IF s < duration(0.0) THEN
    min := min-1;
    s := s+duration(60.0);
  END IF;

  -- compute seconds and hundredth of seconds in sec and sec100
  t := NATURAL(s*100);
  sec := t/100;
  sec100 := t MOD 100;
  RETURN cvis (y MOD 100, 2, '0') & "-" &
    cvis (m, 2, '0') & "-" & cvis (d, 2, '0') & " "
    & cvis (hrs, 2, '0') & ":" & cvis (min, 2, '0') & ":"
    & cvis (sec, 2, '0') & "." & cvis(sec100, 2, '0') ;
END timestamp;

------------------  read and Print procedures -----------------
---------------------------------------------------------------

PROCEDURE Print(v: VARSTRING := outbuffer) IS
BEGIN put_line(v.str(1..v.lg)); END Print;

PROCEDURE Write(v: VARSTRING := outbuffer) IS
BEGIN Put(v.str(1..v.lg)); END Write;

PROCEDURE NewLine(n: INTEGER := 1) IS
BEGIN
  IF n>0 THEN
    new_line(positive_count(n));
  END IF;
END NewLine;

PROCEDURE read IS
BEGIN
  get_line(inbuffer.str, inbuffer.lg);
  IF inbuffer.lg = inbuffer.str'LAST THEN
    set_col(1);
  END IF;
    EXCEPTION WHEN end_error => RAISE enderror;
END read;

--------------------- input from varstrings -------------------------
---------------------------------------------------------------------

FUNCTION GetText (width: INTEGER) RETURN VARSTRING IS
  n: INTEGER := width; m : INTEGER := inbuffer.lg; res: VARSTRING;
BEGIN
  IF n <= 0 THEN
    n := 0;
  END IF;
  IF n > m THEN
    n := m;
  END IF;
  res.lg := n;
  res.str(1 .. n) := inbuffer.str(1..n);
  inbuffer.lg := m-n;
  inbuffer.str(1..inbuffer.lg) :=inbuffer.str(n+1..m);
  RETURN res;
END GetText;

FUNCTION GetInt (width: INTEGER) RETURN INTEGER IS
BEGIN RETURN cvsi(GetText(width)); END GetInt;

FUNCTION GetFloat (width: INTEGER) RETURN FLOAT IS
BEGIN RETURN cvsf(GetText(width)); END GetFloat;


FUNCTION GetText RETURN VARSTRING IS
  n,m,left,right: NATURAL; res: VARSTRING;
BEGIN
  LOOP
    locatetext(left, right);
    EXIT WHEN left <= right;
    read;
  END LOOP;
  n := right-left+1;
  res.str(1 .. n) := inbuffer.str(left .. right);
  m := inbuffer.lg - right;
  inbuffer.str(1..m) := inbuffer.str(right+1..inbuffer.lg);
  res.lg := n;
  inbuffer.lg := m;
  RETURN res;
END GetText;

FUNCTION GetInt RETURN INTEGER IS
BEGIN RETURN cvsi(GetText); END GetInt;

FUNCTION GetFloat RETURN FLOAT IS
BEGIN RETURN cvsf(GetText); END GetFloat;

FUNCTION GetChar RETURN CHARACTER IS
  c: CHARACTER; right : POSITIVE := 1;
BEGIN
  LOOP
    IF right > inbuffer.lg THEN
      read;
      right := 1;
    ELSE
      EXIT WHEN inbuffer.str(right) /= ' ';
      right := right + 1;
   END IF;
  END LOOP;
  c := inbuffer.str(right);
  inbuffer.str(1..inbuffer.lg-right):=inbuffer.str(right+1..inbuffer.lg);
  inbuffer.lg := inbuffer.lg - right;
  RETURN c;
END GetChar;

FUNCTION GetBool RETURN BOOLEAN IS
BEGIN RETURN cvcb(GetChar); END GetBool;

FUNCTION  checkblanks(line: VARSTRING := inbuffer) RETURN BOOLEAN IS
  i,j: NATURAL;
BEGIN locatetext(i,j,line); RETURN (i > j); END checkblanks;

PROCEDURE locatetext (left,right: out INTEGER; line: VARSTRING:= inbuffer) IS
  i: NATURAL := 1; j: NATURAL := 0; -- temporaries for left and right
BEGIN
  WHILE i<=line.lg AND THEN line.str(i)=' ' LOOP
    i := i+1;
  END LOOP;
  j := i;
  WHILE j<=line.lg AND THEN line.str(j)/=' ' LOOP
    j := j+1;
  END LOOP;
  left := i;
  right := j - 1;
END locatetext;


END StringPack;